# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""The main generator function."""

from __future__ import annotations

import dataclasses
import os
from typing import Iterable

from drmock import overload
from drmock import types
from drmock import translator
from drmock import utils

MOCK_OBJECT_PREFIX = 'DRMOCK_Object_'
MOCK_OBJECT_ENCLOSING_NAMESPACE = ('drmock', 'mock_implementation')
METHOD_COLLECTION_NAME = 'methods'
METHOD_CPP_CLASS = 'drmock::Method'
INCLUDE_GUARD_PREFIX = 'DRMOCK_MOCK_IMPLEMENTATIONS_'
DRMOCK_INCLUDE_PATH = 'DrMock/'
MACRO_PREFIX = 'DRMOCK_'
FORWARDING_CTOR_TEMPLATE_PARAMS = MACRO_PREFIX + '_FORWARDING_CTOR_TS'


def main(args) -> None:
    """Generate mock files and save them on disk.

    Args:
        args: Holds the commandline arguments

    The ``args`` parameter is required to have the fields specified in
    the documentation of the ``commandline`` module.

    Raises:
        utils.DrMockRuntimeError:
            If reading/writing any of the specified files fails
    """
    try:
        with open(args.input_path, 'r') as f:
            old_header = f.read()
    except (FileNotFoundError, IOError) as e:
        raise utils.DrMockRuntimeError(str(e))

    macros = {'Q_OBJECT'}
    old_header = _hide_macros_from_preprocessor(old_header, macros)
    new_header, new_source = _main_impl(args, old_header)

    output_path_header = args.output_path
    without_extension, _ = os.path.splitext(output_path_header)
    output_path_source = without_extension + '.cpp'
    try:
        with open(output_path_header, 'w') as f:
            f.write(new_header)
        with open(output_path_source, 'w') as f:
            f.write(new_source)
    except IOError as e:
        raise utils.DrMockRuntimeError(str(e))


def _hide_macros_from_preprocessor(source: str, macros: Iterable[str]) -> str:
    """Hide macros with keywords that drmock recognizes.

    Works by ``#undef``'ing the macro and replacing every occurence by a
    variable declaration.

    Known problems: If ``macros`` is ``['MACRO', 'LONG_MACRO']``, then
    the call will produce unexpected results.
    """
    for each in macros:
        source = source.replace(each, f'#undef {each}\nint {MACRO_PREFIX}{each};')
    return source


def _main_impl(args: str, input_header: str) -> tuple[str, str]:
    """Generate mock header and source code.

    Args:
        args: Holds the commandline arguments
        input_header: The header of the C++ .h to mock

    Returns:
        A tuple which holds the header and source code of the mock class

    Raises:
        utils.DrMockRuntimeError:
            If the clang library file is not set
        utils.DrMockRuntimeError:
            If no class matching the pattern provided by ``args`` is
            found in ``input_header``
    """
    if not args.clang_library_file:
        raise utils.DrMockRuntimeError(
            'clang library file path not set. Specify the path to the clang'
            ' .dll/.so/.dylib using the --clang-library-file command line'
            ' argument or by setting the environment variable'
            ' CLANG_LIBRARY_FILE.')
    translator.set_library_file(args.clang_library_file)
    root = translator.translate(args.input_path, input_header, args.flags)
    node, enclosing_namespace = root.find_matching_class(args.input_class)

    if not node:
        raise utils.DrMockRuntimeError(
            f"No class matching '{args.input_class}' found in {args.input_path}")
    class_ = types.Class.from_node(node)
    class_.enclosing_namespace = enclosing_namespace

    mock_implementation_name = utils.swap(args.input_class, args.output_class, class_.name)

    mock_object = _generate_mock_object(class_, args.access)
    mock_implementation = _generate_mock_implementation(
        mock_implementation_name, class_, args.access)

    new_header = _generate_header(class_, mock_object, mock_implementation, args.input_path)
    new_source = _generate_source(class_, args.output_path)

    return new_header, new_source


def _generate_header(class_: types.Class,
                     mock_object: types.Class,
                     mock_implementation: types.Class,
                     input_path: str) -> str:
    """Generate header code from ``drmock.types.Class`` objects.

    Args:
        class_: The mocked class
        mock_object: The mock object class
        mock_implementation: The mock implementation class
        input_path:
            Absolute or relative path to the mocked class' .h file
    """
    result = ''

    result += _include_guard_open(class_.name)
    result += '\n'

    result += _include_angled_brackets(DRMOCK_INCLUDE_PATH + 'Mock.h')
    result += _include_quotes(os.path.abspath(input_path))
    result += '\n'

    # If explicit instantiations are allowed, declare them in the .h and
    # define them in the .cpp.
    if class_.explicit_instantiation_allowed():
        # Use set in order to discard duplicates (which occur if methods
        # have cv-qualified overloads with the same signature, for example).
        method_templates = {_generate_method_template(class_.full_name(), each)
                            for each in class_.get_virtual_methods()}
        result += '\n'.join(_explicit_instantiation_decl(each) for each in method_templates)
        result += '\n'

    result += str(mock_object)
    result += '\n'
    result += '\n'
    result += str(mock_implementation)
    result += '\n'
    result += '\n'
    result += _include_guard_close(class_.name)

    return result


def _generate_source(class_: types.Class, header_path: str) -> str:
    """Generate mock implementation source code.

    Args:
        class_:
            The mocked class (not the mock object/implementation class!)
        input_path:
            Absolute path to the mock object/implementation .h file
    """
    result = ''
    if class_.explicit_instantiation_allowed():
        # Use set in order to discard duplicates (which occur if methods
        # have cv-qualified overloads with the same signature, for example).
        method_templates = {_generate_method_template(class_.full_name(), each)
                            for each in class_.get_virtual_methods()}
        result += _include_quotes(header_path)
        result += '\n'
        result += '\n'.join(_explicit_instantiation_definition(each)
                            for each in method_templates)
    else:
        result = '// This source file is intentionally left blank'  # To prevent AutoGen warnings.
    return result


def _generate_mock_object(class_: types.Class, access: list[str]) -> types.Class:
    """Generate the ``types.Class`` object of the mock object.

    Args:
        class_:
            The mocked class (not the mock object/implementation class!)
        access: Only mock methods with these access specifiers
    """
    result = types.Class(_generate_mock_object_class_name(class_))
    result.enclosing_namespace = MOCK_OBJECT_ENCLOSING_NAMESPACE
    result.template = class_.template
    overloads = overload.get_overloads_of_class(class_, access)

    type_aliases = class_.get_type_aliases()
    for each in type_aliases:
        each.access = 'private'  # This is for aesthetic reasons!
    result.members += type_aliases

    result.members.append(Friend(class_.full_name(), 'private'))

    # NOTE State object must be default initialized _before_ the
    # shared_ptrs, so it must occur above them in the member list.
    state_object = types.Variable(
        overload.STATE_OBJECT_NAME, 'std::shared_ptr<StateObject>',
        ['std::make_shared<StateObject>()'], access='private')
    result.members.append(state_object)

    shared_ptrs = sum([each.generate_shared_ptrs() for each in overloads], [])
    result.members += shared_ptrs

    method_collection = types.Variable(
        name=METHOD_COLLECTION_NAME,
        type='MethodCollection',
        default_args=['{' + ', '.join(each.name for each in shared_ptrs) + '}'],
        access='private')
    result.members.append(method_collection)

    # NOTE It's important to add the dispatch methods _before_ the
    # getters; otherwise, you'll get the following compiler error:
    #
    # function 'f_dispatch' with # deduced return type cannot be used
    # before it is defined
    result.members += sum([each.generate_dispatch_methods() for each in overloads], [])

    verify_state1 = types.Method(
        name='verifyState',
        return_type='bool',
        params=['const std::string& state'],
        body=f'return {overload.STATE_OBJECT_NAME}->get() == state;')
    verify_state2 = types.Method(
        name='verifyState',
        return_type='bool',
        params=['const std::string& slot', 'const std::string& state'],
        body=f'return {overload.STATE_OBJECT_NAME}->get(slot) == state;')
    result.members.append(verify_state1)
    result.members.append(verify_state2)

    make_formatted_error_string = types.Method(
        name='makeFormattedErrorString',
        return_type='std::string',
        const=True,
        params=[],
        body='return methods.makeFormattedErrorString();')
    result.members.append(make_formatted_error_string)

    verify = types.Method(
        name='verify',
        return_type='bool',
        body=f'return {METHOD_COLLECTION_NAME}.verify();')
    result.members.append(verify)

    result.members += [each.generate_getter() for each in overloads]

    return result


def _generate_mock_implementation(name: str,
                                  class_: types.Class,
                                  access: list[str]) -> types.Class:
    """Generate the ``types.Class`` object of the mock implementation.

    Args:
        name: The name of the mock implementation class
        class_:
            The mocked class (not the mock object/implementation class!)
        access: Only mock methods with these access specifiers
    """
    result = types.Class(name)
    result.enclosing_namespace = class_.enclosing_namespace
    result.template = class_.template
    result.q_object = class_.q_object
    result.parent = class_.full_name()
    result.final = True

    result.members = class_.get_type_aliases()

    # Set the class as parent for all methods.
    overloads = overload.get_overloads_of_class(class_, access)
    default_ctor = types.Constructor(
        name=name,
        template=types.TemplateDecl([f'... {FORWARDING_CTOR_TEMPLATE_PARAMS}']),
        params=[f'{FORWARDING_CTOR_TEMPLATE_PARAMS}&&... ts'],
        initializer_list=[
            result.parent + '{' + f'std::forward<{FORWARDING_CTOR_TEMPLATE_PARAMS}>(ts)...' + '}'],
        body='\n'.join(sum([each.generate_set_parent() for each in overloads], []))
    )
    result.members.append(default_ctor)

    mock_implementation = types.Variable(
        name=overload.MOCK_OBJECT_NAME,
        type=_generate_mock_object_full_class_name(class_),
        mutable=True)
    result.members.append(mock_implementation)

    result.members += sum([each.generate_mock_implementations() for each in overloads], [])

    return result


def _include_quotes(file: str) -> str:
    return f'#include "{file}"\n'


def _include_angled_brackets(file: str) -> str:
    return f'#include <{file}>\n'


def _include_guard_macro(name: str) -> str:
    return INCLUDE_GUARD_PREFIX + name


def _include_guard_open(name: str) -> str:
    result = ''
    result += '#ifndef ' + _include_guard_macro(name) + '\n'
    result += '#define ' + _include_guard_macro(name) + '\n'
    return result


def _include_guard_close(name: str) -> str:
    return '#endif /* ' + _include_guard_macro(name) + ' */'


def _generate_method_template(parent: str, method: types.Method) -> str:
    """Return template for explicit instantiation of C++ Method object.

    Args:
        parent: The parent class' fully qualified name
        method: The target method
    """
    decayed_signature = [str(method.return_type)] \
        + [str(each.get_decayed()) for each in method.params]
    return METHOD_CPP_CLASS + '<' + parent + ', ' + ', '.join(decayed_signature) + '>'


def _explicit_instantiation_decl(expr: str) -> str:
    return 'extern template class ' + expr + ';'


def _explicit_instantiation_definition(expr: str) -> str:
    return 'template class ' + expr + ';'


@dataclasses.dataclass
class Friend:
    name: str
    access: str = 'public'

    def __str__(self):
        return f'friend class {self.name};'


def _generate_mock_object_class_name(class_: types.Class) -> str:
    return MOCK_OBJECT_PREFIX + class_.name


def _generate_mock_object_full_class_name(class_: types.Class):
    """Generate the mock object's name including namespace specifiers
    and template params.

    Args:
        class_:
            The mocked class (not the mock object/implementation class!)
    """
    result = ''
    result += ''.join(each + '::' for each in MOCK_OBJECT_ENCLOSING_NAMESPACE)
    result += _generate_mock_object_class_name(class_)
    if class_.template:
        result += utils.template(class_.template.get_args())
    return result



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""For gathering info about nodes."""

from __future__ import annotations

import functools
from typing import Callable, Optional

import clang.cindex

from drmock import utils

WIDTH = 4


def print_tree(root: translator.Node) -> None:
    def dump_func(node, depth):
        print(_DISPATCH.get(node.cursor.kind, _dump_basic)(node, depth))
    _visit_tree(root, dump_func)


def _visit_tree(root: translator.Node,
                func: Callable[[translator.Node], Any],
                depth: Optional[int] = 0) -> None:
    func(root, depth)
    for each in root.get_children():
        _visit_tree(each, func, depth + 1)


def _indent(func):
    @functools.wraps(func)
    def new_func(node, depth: Optional[int] = 0):
        result = func(node)
        result = utils.indent(result, depth=depth, width=WIDTH)
        return result
    return new_func


@_indent
def _dump_cxx_method(node: translator.Node) -> None:
    cursor = node.cursor
    lines = []
    lines.append(_dump_basic(node))
    tokens = [each.spelling for each in cursor.get_tokens()]
    lines.append(WIDTH * ' ' + 'get_tokens(): ' + str(tokens))
    return '\n'.join(lines)


@_indent
def _dump_type_ref(node: translator.Node) -> None:
    lines = [_dump_basic(node)]
    lines.append(WIDTH * ' ' + 'spelling: ' + str(node.cursor.spelling))
    lines.append(WIDTH * ' ' + 'type.spelling: ' + str(node.cursor.type.spelling))
    return '\n'.join(lines)


@_indent
def _dump_basic(node: translator.Node) -> None:
    return str(node.cursor.kind)


_DISPATCH = {
    clang.cindex.CursorKind.CXX_METHOD: _dump_cxx_method,
    clang.cindex.CursorKind.TYPE_REF: _dump_type_ref,
}



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Commandline client for drmock generator."""

from __future__ import annotations

import argparse
import locale
import textwrap
import os
import subprocess
import sys

from drmock import generator
from drmock import utils

_parser = argparse.ArgumentParser(
    description='Create mock object .h and .cpp files',
    formatter_class=argparse.RawTextHelpFormatter,
    epilog=textwrap.dedent('''
        The .cpp file is saved in the same directory as the .h file.

        The input-class and output-class arguments may be regular
        expressions. input-class must match a class in input-path; if
        multiple matches are found, the first is chosen. output-class
        may contain a backreference (\\1) to a capture group in
        input-class.

        The clang-library-file parameter must either be specified using
        the command line interface, or by setting the CLANG_LIBRARY_FILE
        environment variable.'''))
_parser.add_argument('input_path',
                     help='path to .h file containing the input class')
_parser.add_argument('output_path',
                     help='path to output .h')
# NOTE It's a bit awkward to do the calculation of the mock class'
# name _inside_ the tool, but it's the only place where we have
# access to the mockED class.
_parser.add_argument('--input-class', '-i', default='(.*)',
                     help='name of the input class, default is (.*)')
_parser.add_argument('--output-class', '-o', default=r'Mock\1',
                     help='name of the output/mock class, default is Mock\\1')
# Mock all public virtual functions by default, unless -a=private
_parser.add_argument('--access', '-a', default=['public', 'protected', 'private'],
                     help='only mock virtual functions with the specified access specs')
_parser.add_argument('--flags', '-f', nargs=argparse.REMAINDER, default=[], help='the C++ compiler flags')
# # Mock a selection of virtual functions if -m/--methods=
# _parser.add_argument('--methods', '-m', default=[],
#                      help='only mock specified virtual functions')

_parser.add_argument('--clang-library-file', '-l',
                     default=os.environ.get('CLANG_LIBRARY_FILE', None),
                     help='path to the libclang .dll/.so/.dylib')


def parse_args(args: list[str]) -> argparse.Namespace:
    args = _parser.parse_args(args)

    # Apply isysroot default on macOS.
    if sys.platform == 'darwin' and '-isysroot' not in args.flags:
        tmp = subprocess.check_output(['xcrun', '--show-sdk-path'])
        if sys.stdout.encoding is not None:
            encoding = sys.stdout.encoding
        else:
            encoding = locale.getpreferredencoding()
        sdk = tmp.decode(encoding).rstrip('\n')
        args.flags.append('-isysroot')
        args.flags.append(sdk)

    return args


# This method is the entry point of the drmock-gen script.
def main() -> None:
    try:
        args = parse_args(sys.argv[1:])
        # Due to the way that argparse parses args, the need to strip
        # the first compiler flag of whitespace!
        if args.flags:
            args.flags[0] = args.flags[0].lstrip()
        generator.main(args)
    except utils.DrMockRuntimeError as e:  # FIXME _Don't_ print traceback on clang errors, etc.!
        print(f'drmock-gen: error: {e}\n', file=sys.stderr)
        sys.exit(1)



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""For mocking overloaded C++ methods."""

from __future__ import annotations

import copy
import dataclasses
from typing import Iterator, Sequence

from drmock import types
from drmock import utils

MOCK_OBJECT_NAME = 'mock'
SHARED_PTR_PREFIX = 'METHODS_DRMOCK_'
STATE_OBJECT_NAME = 'STATE_OBJECT_DRMOCK_'
CONST_ENUM = 'drmock::Const'
VOLATILE_ENUM = 'drmock::Volatile'
TYPE_CONTAINER = 'TypeContainer'
PARAMETER_PACK = '... DRMOCK_Ts'


def get_overloads_of_class(class_: types.Class,
                           access_specs: Iterator[str] = None) -> list[Overload]:
    """Group method of ``class_`` into ``Overload`` objects.

    Args:
        class_: The class whose methods are grouped
        acccess_specs: Only group method with these access specifiers

    Returns:
        A list with the ``Overload`` objects
    """
    if not access_specs:
        access_specs = ['public']
    virtual_methods = [each for each in class_.get_virtual_methods()
                       if each.access in access_specs]
    collections = utils.split_by_condition(lambda f: f.mangled_name(), virtual_methods)
    return [Overload(class_, each) for each in collections]


class Overload:
    """Represents a C++ method overload.

    All methods are assumed to be _non-template_!
    """

    def __init__(self, parent: types.Class, methods: Sequence[types.Method]) -> None:
        self._parent = parent
        self._methods = methods

    def is_overload(self) -> bool:
        """Check if ``self`` is a proper overload (with at least two
        methods)."""
        return len(self._methods) > 1

    def generate_getter(self) -> types.Method:
        """Generate the overload's template getter method."""
        f = self._methods[0]  # Representative of the overload.
        result = types.Method('f')  # Use temporary dummy name for initialization!
        result.template = types.TemplateDecl([PARAMETER_PACK])
        result.name = f.mangled_name()
        result.return_type = types.Type.from_spelling('auto &')

        # Create arguments for dispatch call.
        dispatch = result.template.get_args()

        # If all methods have the same parameter types, then these must
        # automatically be passed as arguments to the dispatch call.
        # NOTE When using strings, spelling differences between equal
        # types (``T *`` vs. ``T*``) can cause this part to malfunction.
        if not self.is_overload():  # all(f.params == each.params for each in self._methods):
            dispatch.extend(f.params)

        # If all methods are const qualified, then the const qualifier
        # must automatically be passed to the dispatch method.
        if all(each.const for each in self._methods):
            dispatch.append(CONST_ENUM)

        result.body = f'return {f.mangled_name()}_dispatch(' + TYPE_CONTAINER + \
            utils.template(dispatch) + '{});'
        result.access = 'public'
        return result

    def generate_shared_ptrs(self) -> list[types.Variable]:
        """Generate the overload's ``shared_ptr<Method>`` objects."""
        result = []
        for i, f in enumerate(self._methods):
            template_args = [self._parent.full_name(), f.return_type]
            template_args.extend(each.get_decayed() for each in f.params)

            value_type = 'Method' + utils.template(template_args)
            ptr = f'std::make_shared<{value_type}>("", {STATE_OBJECT_NAME})'
            shared_ptr = types.Variable(
                name=_shared_ptr_name(f.mangled_name(), i),
                type=f'std::shared_ptr<{value_type}>',
                default_args=[ptr],
                access='private')
            result.append(shared_ptr)
        return result

    def generate_dispatch_methods(self) -> list[types.Method]:
        """Generate the overload's dispatch methods."""
        result = []
        for i, f in enumerate(self._methods):
            dispatch = types.Method('f')  # Temporary name for init.
            dispatch.name = f.mangled_name() + '_dispatch'
            dispatch.return_type = types.Type('auto', lvalue_ref=True)

            # The method's cv qualifiers are stored in the type
            # container's template args, together with the types of the
            # params of ``f``.
            template_args = copy.deepcopy(f.params)
            if f.const:
                template_args.append(types.Type(CONST_ENUM))
            if f.volatile:
                template_args.append(types.Type(VOLATILE_ENUM))
            dispatch.params = [types.Type(TYPE_CONTAINER + utils.template(template_args))]
            dispatch.body = 'return *' + _shared_ptr_name(f.mangled_name(), i) + ';'
            dispatch.access = 'private'
            result.append(dispatch)
        return result

    def generate_mock_implementations(self) -> list[types.Method]:
        """Generate the mock implementations of the overload's
        functions."""
        result = []
        for f in self._methods:
            impl = copy.deepcopy(f)  # Keep name, cv qualifiers.
            impl.virtual = False
            impl.pure_virtual = False
            impl.override = True
            impl.params = [f'{each} a{i}' for i, each in enumerate(f.params)]  # (Add parameter names!) # noqa: E501

            # If the method is not overloaded, the correct template
            # arguments are automatically used, and need not be manually
            # inserted.
            dispatch_call = _generate_access(f, self.is_overload())
            dispatch_call += '.call(' + ', '.join(
                _unpack_and_move(f'a{i}', each) for i, each in enumerate(f.params)) + ');'

            # If ``f`` is non-void, then the result of the call must be
            # forwarded.
            impl.body = ''  # Reset body!
            if f.return_type == types.Type('void'):
                impl.body += dispatch_call
            else:
                impl.body += f'auto& result = *{dispatch_call}\n'
                impl.body += f'return std::forward<{f.return_type}>(drmock::moveIfNotCopyConstructible(result));'  # noqa: E501

            impl.access = 'public'

            result.append(impl)

        return result

    def generate_set_parent(self) -> list[str]:
        """For each function, return a C++ statement to set the parent
        class."""
        return [_generate_access(each, self.is_overload()) + '.parent(this);'
                for each in self._methods]


def _shared_ptr_name(mangled_name: str, i: int) -> str:
    return f'{SHARED_PTR_PREFIX}{mangled_name}_{i}'


def _generate_access(f: types.Method, overload: bool) -> str:
    """Return code for accessing method ptr from mock object.

    Args:
        f: The method object
        overload: True if ``f`` is overloaded
    """
    if overload:
        template_args = copy.deepcopy(f.params)
        if f.const:
            template_args.append(CONST_ENUM)
        if f.volatile:
            template_args.append(VOLATILE_ENUM)
        result = MOCK_OBJECT_NAME + '.template ' + f.mangled_name() \
            + utils.template(template_args) + '()'
    else:
        result = MOCK_OBJECT_NAME + '.' + f.mangled_name() + '()'
    return result


def _unpack_and_move(name: str, type_: types.Type):
    result = name if type_.lvalue_ref else f'std::move({name})'
    result += type_.parameter_pack * '...'
    return result



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""libclang wrapper module.

This module translates a C++ source into the corresponding AST using the
``translate``. Before calling ``translate``, you must set the path to
the ``libclang.dll/.so/.dylib`` _file_ using ``set_library_file``.
"""

from __future__ import annotations

import re
from typing import Optional

import clang.cindex

from drmock import utils

DIAGNOSTIC_FORMAT_OPTIONS = (clang.cindex.Diagnostic.DisplaySourceLocation
                           | clang.cindex.Diagnostic.DisplayColumn
                           | clang.cindex.Diagnostic.DisplaySourceRanges
                           | clang.cindex.Diagnostic.DisplayOption
                           | clang.cindex.Diagnostic.DisplayCategoryId
                           | clang.cindex.Diagnostic.DisplayCategoryName)

CLASS_CURSORS = {clang.cindex.CursorKind.CLASS_DECL, clang.cindex.CursorKind.CLASS_TEMPLATE}


def set_library_file(file: str) -> None:
    """Args:
        file: path to libclang dynamic library.
    """
    clang.cindex.Config.set_library_file(file)


class Node:
    """Wrapper class for ``clang.cindex.Cursor`` which tracks file
    membership."""

    def __init__(self, cursor: clang.cindex.Cursor, path: str) -> None:
        """Args:
            cursor: The wrapper cursor
            path: The file that the cursor belongs to
        """
        self._cursor = cursor
        self._path = path

    @property
    def cursor(self) -> clang.cindex.Cursor:
        return self._cursor

    def get_children(self) -> list[Node]:
        """Get all children from the same file."""
        return [Node(each, self._path) for each in self._cursor.get_children()
                if str(each.location.file) == self._path]

    def get_tokens(self) -> list[str]:
        """Get the cursor's tokens."""
        return [each.spelling for each in self._cursor.get_tokens()]

    def find_matching_class(self,
                            regex: str,
                            enclosing_namespace: Optional[list[str]] = None
                            ) -> tuple[Optional[Node], list[str]]:
        """Search the tree under ``self`` for a class whose name matches
        ``regex``.

        Args:
            regex: The regex to match the sought class' name against
            enclosing_namespace:
                Only used for recursing, **must** not be set by the user
        """
        if not enclosing_namespace:
            enclosing_namespace = []

        for each in self.get_children():
            if each.cursor.kind == clang.cindex.CursorKind.NAMESPACE:
                enclosing_namespace.append(each.cursor.displayname)
                return each.find_matching_class(regex, enclosing_namespace)
                enclosing_namespace.pop()  # Remove namespace upon leaving the node!
            if each.cursor.kind in CLASS_CURSORS and re.match(regex, each.cursor.spelling):
                return each, enclosing_namespace
        return None, []


def translate_file(path: str, compiler_flags: Optional[list[str]] = None) -> Node:
    """Translate the content of ``path`` into its AST.

    Args:
        path: The path to the file
        compiler_flags: A list of compiler flags used for parsing

    Raises:
        clang.cindex.LibclangError:
            If the libclang path is not set or not found (see
            ``set_library_file``)
        utils.DrMockRuntimeError:
            If ``path`` is empty
    """
    with open(path, 'r') as f:
        source = f.read()
    return translate(path, source, compiler_flags)


def translate(path: str, source: str, compiler_flags: Optional[list[str]] = None) -> Node:
    """Translate a string with C++ code into its AST.

    Args:
        path: The path of the parsed file
        source: The C++ source
        compiler_flags: A list of compiler flags used for parsing

    Raises:
        clang.cindex.LibclangError:
            If the libclang path is not set or not found (see
            ``set_library_file``)
        utils.DrMockRuntimeError:
            If ``path`` is empty

    Note: The ``path`` parameter is required due to ``clang`` details.
    It need not be a real path, but it must be non-empty. Choosing a
    unique name is useful, as it is used in clang's diagnostics.
    """
    if not path:
        raise utils.DrMockRuntimeError(
            "translate failed: Parameter 'path' is empty. Expected non-empty string")

    if compiler_flags is None:
        compiler_flags = []

    index = clang.cindex.Index.create()
    try:
        tu = index.parse(path, ['-x', 'c++'] + compiler_flags, unsaved_files=[(path, source)])
    except clang.cindex.TranslationUnitLoadError as e:
        raise utils.DrMockRuntimeError(str(e))

    # Check for errors.
    if tu.diagnostics:
        error = 'Clang failed. Details:\n\n'
        error += '\n'.join('\t' + each.format(DIAGNOSTIC_FORMAT_OPTIONS)
                           for each in tu.diagnostics)
        raise utils.DrMockRuntimeError(error)

    return Node(tu.cursor, path)



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Classes for C++ AST node infos.

Most classes represent a type of AST node and feature a ``from_node``
``classmethod`` which may be created from an appropriate
``translator.Node`` object:

```python
@classmethod
def from_node(cls: _T, node: translator.Node) -> _T:
    ...
```

The ``node`` parameter **must** have the correct cursor kind, otherwise
the call will fail with an error.
"""

from __future__ import annotations

import copy
import collections
import dataclasses
from typing import Any, Optional, Sequence, Union
import clang.cindex

from drmock import utils

"""We're using an ``OrderedDict`` to ensure that in ``Method.mangled_name``
for example ``<=>`` is replaced _before_ ``<=`` to ensure that
``operator<=>`` becomes ``operatorSpaceship`` instead of
``operatorLesserOrEqualGreater``.
"""
_OPERATOR_SYMBOLS = collections.OrderedDict([
    ('<=>', 'SpaceShip'),
    ('->*', 'PointerToMember'),
    ('co_await', 'CoAwait'),
    ('==', 'Equal'),
    ('!=', 'NotEqual'),
    ('<=', 'LesserOrEqual'),
    ('>=', 'GreaterOrEqual'),
    ('<<', 'StreamLeft'),
    ('>>', 'StreamRight'),
    ('&&', 'And'),
    ('||', 'Or'),
    ('++', 'Increment'),
    ('--', 'Decrement'),
    ('->', 'Arrow'),
    ('()', 'Call'),
    ('[]', 'Brackets'),
    ('+', 'Plus'),
    ('-', 'Minus'),
    ('*', 'Ast'),
    ('/', 'Div'),
    ('%', 'Modulo'),
    ('^', 'Caret'),
    ('&', 'Amp'),
    ('|', 'Pipe'),
    ('~', 'Tilde'),
    ('!', 'Not'),
    ('=', 'Assign'),
    ('<', 'Lesser'),
    ('>', 'Greater'),
    (',', 'Comma'),
])


def from_node(node: translator.Node) -> Any:
    """Create an instance of an appropriate class of this module from a
    node.

    Args:
        The node to the create the object from

    The type of the instance is determined from the node's ``kind``
    property.

    Raises:
        ValueError: If the appropriate class cannot be determined
    """
    if not hasattr(from_node, '_DISPATCH'):
        from_node._DISPATCH = {
            clang.cindex.CursorKind.PARM_DECL: Type,
            clang.cindex.CursorKind.CXX_METHOD: Method,
            clang.cindex.CursorKind.TYPE_ALIAS_TEMPLATE_DECL: TypeAlias,
            clang.cindex.CursorKind.TYPE_ALIAS_DECL: TypeAlias,
        }
    cursor = node.cursor
    type_ = from_node._DISPATCH.get(cursor.kind, None)
    if not type_:
        keys = ', '.join(each.cursor.kind.name for each in from_node._DISPATCH)
        raise ValueError(f'Invalid CursorKind. Expected: {keys}; received: {cursor.kind.name}')
    return type_.from_node(node)


@dataclasses.dataclass
class Type:
    """For C++ type declarations.

    Types are stored in "layers", where each layer represents an
    indirection due to cv qualifiers, references or pointers. Each layer
    points to the next layer using the ``inner`` attribute. The
    inner-most layer is a string which holds the spelling of the core
    type (for example, the inner-most type for  ``const
    std::vector<int>&`` is ``'std::vector<int>'``).

    Due to the nature of this design, artificial _naked_ layers which
    hold no information are possible. For example:

    >>> type_ = Type('int')
    >>> naked = Type(type_)  # Represents same type, but with naked first layer
    """
    inner: Union[str, Type]
    const: bool = False
    volatile: bool = False
    lvalue_ref: bool = False
    rvalue_ref: bool = False
    pointer: bool = False
    parameter_pack: bool = False

    def get_decayed(self) -> Type:
        """Return the decayed version of ``self``."""
        # Note that if an instance of `Type` is a reference, then its
        # const qualifier is saved in the inner type. This is due to the
        # confusion between the terms "const reference" and "reference
        # to const": The distinction between "const reference" and
        # "reference to const" is unnecessary, and the former is only a
        # shorthand for the latter. The same goes for volatile
        # qualifiers. Therefore, when decaying a reference, the cv
        # qualifiers must be removed from the inner type; otherwise,
        # they must be removed from the instance itself.

        result = self._get_simplified()  # In case top item is naked.
        if result.lvalue_ref or result.rvalue_ref:
            result.lvalue_ref = False
            result.rvalue_ref = False
            if not isinstance(result.inner, str):
                result.inner.const = False
                result.inner.volatile = False
        else:
            result.const = False
            result.volatile = False
        result = result._get_simplified()  # If ``result`` was a reference, it's now naked.
        return result

    def _get_simplified(self, first_pass: bool = True) -> Type:
        """Return equivalent type with naked layers removed.

        Args:
            first_pass:
                Optional argument for recursing, **must** not be set by
                the caller
        """
        # Find outer-most non-naked position in the hierarchy. This type
        # will form the outer Type object of the result.
        result = self
        while not isinstance(result.inner, str) and result._is_naked():
            result = result.inner

        # Create a copy of the new base object on the first iteration.
        # On later iterations, we need to make sure that
        if first_pass:
            result = copy.deepcopy(result)

        # If the outer type is a base-type, we're done. On a later
        # iteration, the last type may be naked and must be skipped. On
        # the first iteration, the last naked type is used as wrapper.
        if isinstance(result.inner, str):
            if not first_pass and result._is_naked():
                return result.inner
            return result

        # Otherwise, simplify the inner type.
        result.inner = result.inner._get_simplified(False)
        return result

    def _is_naked(self) -> bool:
        """Check if ``self`` is naked."""
        return (not self.const and not self.volatile and not self.lvalue_ref and
                not self.rvalue_ref and not self.pointer and not self.parameter_pack)

    def __str__(self):
        result = str(self.inner)
        result += self.lvalue_ref * ' &' + self.rvalue_ref * ' &&'
        if self.pointer:
            result += ' *' + self.const * ' const' + self.volatile * ' volatile'
        else:
            result = self.const * 'const ' + self.volatile * 'volatile ' + result
        result += self.parameter_pack * ' ...'
        return result

    @classmethod
    def from_node(cls, node: translator.Node) -> Type:
        # NOTE The following is a hack to solve some rather unfortunate
        # behavior of python clang. When using a type alias such as
        #
        # namespace outer { namespace inner {
        #
        # class Foo
        # {
        # public:
        #   using T = std::shared_ptr<int>;
        # }
        #
        # } // namespace outer::inner
        #
        # in a function declaration,
        #
        # void f(T);
        #
        # ``T`` is expanded into ``inner::outer::Foo``, unless ``T`` is
        # suffciently buried (for instance, ``std::shared_ptr<T>`` will
        # not be expanded). This seems to happen with ``spelling``,
        # ``type.spelling``, ``displayname``, etc.
        #
        # Leaving this unchanged would render the ``Method`` object
        # corresponding to this declaration dependent on the class that
        # the declaration occured in. But we want to move the method (or
        # the type alias) into another class!
        #
        # Another matter is that, even if ``Foo`` is a class template, the
        # expanded name will not contain the template parameters
        # (``outer::inner::Foo`` instead of ``outer::inner::Foo<...>``).
        # Therefore, parsing and printing a class template with type
        # aliases will result in code that will raise a compiler error.
        #
        # The problem is solved by taking the tokens of the parameter
        # declarations and joining them into a string.
        #
        # But two related problems arise when using tokens. If a
        # variable name is given to a parameter, as in
        #
        # f(const T & ... foo);
        #                 ^^^
        #
        # then that name will appear in the tokens. It must therefore be
        # removed from the tokens. But beware! If no variable name or
        # '...' is present in the parameter declaration, then an outer
        # ``const`` as in
        #
        # f(const T* const);
        #            ^^^^^
        #
        # is lost when considering the cursor's tokens, and thus
        # ``from_tokens`` cannot be used. Fortunately, this outer const
        # qualifier can be recognized using clang python. Thus, in the
        # case of a parameter pack or a parameter decl with variable
        # name, ``from_tokens`` may be called after popping the variable
        # name. Otherwise, a complex route must be taken.

        # Check for variable name or parameter pack and call ``from_tokens``.
        tokens = node.get_tokens()  # ['const', 'T', '&', '...', 'foo']
        var = node.cursor.spelling  # 'foo'
        if var != '' and tokens[-1] == var:  # Remove variable/parameter name.
            tokens.pop()
        result = cls.from_tokens(tokens)
        # In some cases (e.g. ``const T*const``), the outer const is not
        # found in the tokens, so we must use class methods.
        result.const = node.cursor.type.is_const_qualified()
        result.volatile = node.cursor.type.is_volatile_qualified()
        return result

    @classmethod
    def from_tokens(cls, tokens: Sequence[str]) -> Type:
        """Create a ``Type`` instance from a sequence of tokens."""
        t = cls('T')  # Use temporary inner name to init ``t``.

        # Read from the right.
        while True:
            if tokens[-1] == 'const':
                t.const = True
                tokens.pop()
            elif tokens[-1] == 'volatile':
                t.volatile = True
                tokens.pop()
            elif tokens[-1] == '...':
                t.parameter_pack = True
                tokens.pop()
            elif tokens[-1] == '*':
                t.pointer = True
                tokens.pop()
                break
            elif tokens[-1] == '&':
                t.lvalue_ref = True
                tokens.pop()
                break
            elif tokens[-1] == '&&':
                t.rvalue_ref = True
                tokens.pop()
                break
            else:
                break

        # If ``t`` is a pointer or a reference, reassemble the remaining
        # tokens and call ``from_tokens`` recursively.
        if t.pointer or t.lvalue_ref or t.rvalue_ref:
            t.inner = Type.from_tokens(tokens)
            return t._get_simplified()

        # If ``t`` is not a pointer or a reference, then read from the
        # left.
        while True:
            if tokens[0] == 'const':
                t.const = True
                tokens.pop(0)
            elif tokens[0] == 'volatile':
                t.volatile = True
                tokens.pop(0)
            else:
                break

        # Terminate the recursion by reassembeling the remaining tokens.
        t.inner = ' '.join(tokens)
        return t._get_simplified()

    @classmethod
    def from_spelling(cls, spelling: str) -> Type:
        """Create a ``Type`` instance from a cursor spelling."""
        tokens = spelling.split(' ')
        return cls.from_tokens(tokens)


class TemplateDecl:  # For TemplateDeclaration
    """For template declarations."""

    def __init__(self, params: Sequence[str]):
        """Args:
            params: A list of the template decl's params

        The ``class``/``typename`` keyword **must** be omitted. Variadic
        template parameters are specified using a prefixed ``'... '``.
        For example:

        >>> decl = TemplateDecl(['T', '... Ts'])
        """
        self._params = params

    def get_args(self) -> list[str]:
        """Get the decl's args.

        Note that the tokens of a variadic template param are swapped.
        For example:

        >>> decl = TemplateDecl(['T', '... Ts'])
        >>> decl.get_args()
        ['T', 'Ts ...']
        """
        return [utils.swap(r'\.\.\. (.*)', r'\1 ...', each)
                if each.startswith('...') else each
                for each in self._params]

    def __eq__(self, other):
        if not isinstance(other, TemplateDecl):
            return NotImplemented
        return self._params == other._params

    def __str__(self):
        result = ''
        result += 'template<'
        result += ', '.join(f'typename {each}' for each in self._params)
        result += '>'
        return result

    @classmethod
    def from_node(cls, node: translator.Node) -> TemplateDecl:
        params = []
        for each in node.get_children():
            if each.cursor.kind == clang.cindex.CursorKind.TEMPLATE_TYPE_PARAMETER:
                # The following is a hack used to circumvent the problem
                # that `cindex.CursorKind.TEMPLATE_TYPE_PARAMETER` is
                # unable to recognize variadic template parameters.
                # Fortunately, it is possible to detect "..." using the
                # `get_tokens()` method.
                tokens = each.get_tokens()  # ["typename", "T"] or ["typename", "...", "Ts"]
                name = ' '.join(tokens[1:])  # "T" or "... Ts"
                params.append(name)
        result = TemplateDecl(params)
        return result


class Constructor:
    """For C++ class constructors."""

    def __init__(self,
                 name: str,
                 params: Optional[Sequence[Union[str, Type]]] = None,
                 template: Optional[TemplateDecl] = None,
                 initializer_list: Optional[Sequence[str]] = None,
                 body: str = '',
                 access: str = 'public'):
        """Args:
            name: The ctor's name
            params: The params
            template: A template decl if the ctor is templated
            initializer_list: The entries of the ctor's initializer list
            body: The body
            access: The access specifier of the object
        """
        self._name = name
        if params is not None:
            self._params = params
        else:
            self._params = []
        self._template = template
        if initializer_list is not None:
            self._initializer_list = initializer_list
        else:
            self._initializer_list = []
        self._body = body
        self._access = access

    @property
    def access(self) -> str:
        """The access specifier of the object."""
        return self._access

    def __str__(self) -> str:
        result = ''
        if self._template:
            result += str(self._template) + '\n'
        result += self._name
        params = ', '.join(str(each) for each in self._params)
        result += f'({params})'
        if self._initializer_list:
            result += ' : ' + ', '.join(self._initializer_list)
        result += '\n{\n'
        result += utils.indent(self._body)
        result += '\n}'
        return result


@dataclasses.dataclass
class Method:
    """For C++ class methods.

    Attributes:
        name: The method's name
        params: The method's parameters
        return_type: The method's return type
        template: The method's template declaration (if available)
        const: Indicates if the method is const-qualified
        volatile: Indicates if the method is volatile-qualified
        virtual: Indicates if the method is virtual
        pure_virtual: Indicates if the method is pure virtual
        override: Indicates if the method has the ``override`` specifier
        noexcept: Indicates if the method has the ``noexcept`` specifier
        operator: Indicates if the method is an operator
        body: The method's body (if available)
        access: The method's access specifier

    Beware! By manually setting these attributes, you **may** impossible
    configurations, methods which are pure virtual but not virtual, for
    example.

    Each method has a _mangled name_. If the method is an operator, the
    operator symbol is replaced with an alphanumeric string which
    describes that symbol (see ``_OPERATOR_SYMBOLS``), otherwise, the
    mangled name is the method's normal name.
    """
    name: str
    params: Sequence[Union[str, Type]] = dataclasses.field(default_factory=list)
    return_type: Type = Type('void')
    template: Optional[TemplateDecl] = None
    const: bool = False
    volatile: bool = False
    virtual: bool = False
    pure_virtual: bool = False
    override: bool = False
    noexcept: bool = False
    operator: bool = False
    body: Optional[str] = None
    access: str = 'public'

    def mangled_name(self) -> str:
        """Return the mangled name of the method."""
        result = self.name
        for k, v in _OPERATOR_SYMBOLS.items():
            result = result.replace(k, v)
        return result

    def __str__(self) -> str:
        result = ''
        if self.template:
            result += str(self.template) + '\n'
        if self.virtual:
            result += 'virtual '
        if self.return_type:  # This check makes sure that no ugly indentation occurs for ctors!
            result += str(self.return_type) + ' '
        result += self.name
        params = ', '.join(str(each) for each in self.params)
        result += f'({params})'
        result += self.const * ' const'
        result += self.volatile * ' volatile'
        result += self.noexcept * ' noexcept'
        result += self.override * ' override'  # This must always be last!
        if self.body:
            result += '\n{\n'
            result += utils.indent(self.body)
            result += '\n}'
        else:
            if self.pure_virtual:
                result += ' = 0'
            result += ';'
        return result

    @classmethod
    def from_node(cls, node: translator.Node) -> Method:
        # NOTE The following is a hack to solve some rather unfortunate
        # behavior of python clang. When using a type alias such as
        #
        # namespace outer { namespace inner {
        #
        # class Foo
        # {
        # public:
        #   using T = std::shared_ptr<int>;
        # }
        #
        # } // namespace outer::inner
        #
        # in a function declaration,
        #
        # void f(T);
        #
        # ``T`` is expanded into ``inner::outer::Foo``, unless ``T`` is
        # suffciently buried (for instance, ``std::shared_ptr<T>`` will
        # not be expanded). This seems to happen with ``spelling``,
        # ``type.spelling``, ``displayname``, etc.
        #
        # Leaving this unchanged would render the ``Method`` object
        # corresponding to this declaration dependent on the class that
        # the declaration occured in. But we want to move the method (or
        # the type alias) into another class!
        #
        # Another matter is that, even if ``Foo`` is a class template, the
        # expanded name will not contain the template parameters
        # (``outer::inner::Foo`` instead of ``outer::inner::Foo<...>``).
        # Therefore, parsing and printing a class template with type
        # aliases will result in code that will raise a compiler error.
        #
        # ``Method.from_node()`` uses ``cindex.Type.get_canonical()`` to
        # assemble the return type. If ``Foo`` is a class template and a
        # template parameter of ``Foo`` appears as parameter in the type
        # alias, it is represented in the form
        #
        # shared_ptr<type-parameter-0-i>
        #
        # where ``i`` is the index of the parameter (note the missing
        # ``std::``).
        #
        # The current solution is to gather the tokens of the CXX_METHOD
        # cursor between the virtual keyword and the function name, and
        # pass these to ``Type.from_tokens``.
        #
        # virtual const T * const f() = 0;
        #         ^^^^^^^^^^^^^^^
        #
        # Special care must be taken when dealing with operators.

        f = Method(node.cursor.spelling)
        f.params = [from_node(each) for each in node.get_children()
                    if each.cursor.kind == clang.cindex.CursorKind.PARM_DECL]
        tokens = node.get_tokens()
        if tokens[0] == 'virtual':
            tokens.pop(0)

        # Check if the name of ``f`` occurs in the tokens. If not, then
        # ``f`` is an operator.
        if f.name not in tokens:
            delim = tokens.index('operator')
        else:
            delim = tokens.index(f.name)
        f.return_type = Type.from_tokens(tokens[:delim])

        # The following is a hack to obtain the volatile qualifier and
        # override keywords:
        #
        # void f(...) const volatile noexcept = 0;
        #             ^^^^^^^^^^^^^^^^^^^^^^^
        #
        # Get these tokens! (There is, apparently, no other way to check
        # for cv qualifiers using python clang.)
        delim = max(index for index, value in enumerate(tokens)
                    if value == ')')  # Index of last closing parens.
        keywords = tokens[delim + 1:]
        if 'override' in keywords:
            f.override = True
        if 'volatile' in keywords:
            f.volatile = True

        # Const qualifiers, virtual keywords, and exception
        # specifications can be obtained using python clang.
        f.const = node.cursor.is_const_method()
        f.virtual = node.cursor.is_virtual_method()
        f.pure_virtual = node.cursor.is_pure_virtual_method()
        if (node.cursor.exception_specification_kind == clang.cindex.ExceptionSpecificationKind.BASIC_NOEXCEPT):  # noqa: E501
            f.noexcept = True

        # ``f`` is an operator, if its name matches the following regex.
        if f.name.replace('operator', '') in _OPERATOR_SYMBOLS:
            f.operator = True
        return f


@dataclasses.dataclass
class TypeAlias:
    """For C++ (template and non-template) type aliases.

    Attributes:
        name: The alias
        typedef: The aliased type
        template: The type alias' template decl (if available)

    Example:
        >>> TypeAlias('Vector', 'std::vector<T>', TemplateDecl(['T']))
        # template<typename T> using Vector<T> = std::vector<T>;
    """
    name: str
    typedef: str
    template: Optional[TemplateDecl] = None

    def __str__(self):
        result = ''
        if self.template:
            result += str(self.template) + ' '
        result += f'using {self.name} = {str(self.typedef)};'
        return result

    @classmethod
    def from_node(cls, node: translator.Node) -> TypeAlias:
        if node.cursor.kind == clang.cindex.CursorKind.TYPE_ALIAS_TEMPLATE_DECL:
            template = TemplateDecl.from_node(node)
            node = next(each for each in node.get_children()
                        if each.cursor.kind == clang.cindex.CursorKind.TYPE_ALIAS_DECL)
        else:
            template = None
        result = TypeAlias(node.cursor.spelling, node.cursor.underlying_typedef_type.spelling)
        result.template = template
        return result


@dataclasses.dataclass
class Class:
    """For C++ classes.

    Attributes:
        name: The class name
        enclosing_namespace: The namespace the class is contained in
        members: A list of the class' members (see note below)
        final: Indicates if the class is final
        q_object: Indicates if the class is a ``Q_OBJECT``
        parent: The name of the class' parent (if available)
        template: The template decl (if available)

    Note that this class is not a faithful representation of C++
    classes. For example, the order of members may change during
    loading. Also, we don't check for base classes, etc.
    """
    name: str
    enclosing_namespace: list[str] = dataclasses.field(default_factory=list)
    members: list[Union[Constructor, Method, Variable]] = dataclasses.field(
        default_factory=list)  # Methods, variables, type aliases!
    final: bool = False
    q_object: bool = False
    # We're assuming at most one (public) base class! NOTE This does not mean
    # that the mockee cannot inherit from more than one parent. We only use
    # the ``parent`` field for later printing the mock class.
    parent: Optional[str] = None
    template: Optional[TemplateDecl] = None

    def full_name(self) -> str:
        """Return the fully qualified class name."""
        result = ''
        result += ''.join(each + '::' for each in self.enclosing_namespace)
        result += self.name
        if self.template:
            result += utils.template(self.template.get_args())
        return result

    def get_virtual_methods(self) -> list[Method]:
        return [each for each in self.members if isinstance(each, Method) and each.virtual]

    def get_type_aliases(self) -> list[TypeAlias]:
        return [each for each in self.members if isinstance(each, TypeAlias)]

    def explicit_instantiation_allowed(self) -> bool:
        """Check if explicit instantiations of the class' method are
        possible.

        As of C++17, this means that the class is not a class template
        nor contains type aliases.
        """
        return (self.template is None) and not self.get_type_aliases()

    def __str__(self):
        result = ''

        if self.enclosing_namespace:
            result += ' '.join('namespace ' + each + ' {' for each in self.enclosing_namespace)
            result += '\n'
            result += '\n'

        if self.template:
            result += str(self.template) + '\n'

        result += 'class ' + self.name + self.final * ' final'
        if self.parent:  # If self is derived (as C++ class).
            result += ' : public ' + self.parent

        result += '\n'  # End class decl line.
        result += '{\n'

        result += self.q_object * '  Q_OBJECT\n\n'

        access = 'private'
        for each in self.members:
            # Observe access specifier change.
            if access != each.access:
                access = each.access
                result += '\n'
                result += each.access + ':\n'

            result += utils.indent(str(each))
            result += '\n'

        result += '};'

        if self.enclosing_namespace:
            result += '\n\n'
            namespace_count = len(self.enclosing_namespace)
            result += namespace_count * '}' + ' // namespace ' + \
                '::'.join(self.enclosing_namespace)

        return result

    @classmethod
    def from_node(cls, node: translator.Node) -> Class:
        """Create ``Class`` object from node.

        Args:
            node: The node to create the object from

        Returns:
            The newly created ``Class`` object

        Note that the ``enclosing_namespace`` field of the ``Class``
        object will *not* be set, as they cannot be read from the class
        node. This is done by the
        ``translator.Node.find_matching_class`` method.

        Note that ``from_node`` will only transcribe information
        relevant to our mocking process and therefore not faithfully
        represent the underlying C++ class. For example, ctor, dtors,
        field variables, etc. will *not* be transcribed into the
        ``Class`` object.
        """
        assert node.cursor.kind in {
            clang.cindex.CursorKind.CLASS_DECL, clang.cindex.CursorKind.CLASS_TEMPLATE}
        result = cls('T')  # Use temporary class name for init.
        result.name = node.cursor.spelling
        if node.cursor.kind == clang.cindex.CursorKind.CLASS_TEMPLATE:
            result.template = TemplateDecl.from_node(node)

        access = 'private'
        for each in node.get_children():
            IGNORED_CURSORS = {
                clang.cindex.CursorKind.TEMPLATE_TYPE_PARAMETER,
                clang.cindex.CursorKind.DESTRUCTOR,
                clang.cindex.CursorKind.CXX_BASE_SPECIFIER,
                clang.cindex.CursorKind.USING_DIRECTIVE,
            }
            if each.cursor.kind in IGNORED_CURSORS:
                continue

            # A field decl can mean one of two things: (1) A field in a
            # non-abstract base class, (2) a macro notification created
            # by us.
            if each.cursor.kind == clang.cindex.CursorKind.FIELD_DECL:
                tokens = each.get_tokens()  # ['int', 'Q_OBJECT'], etc.
                assert tokens
                tokens.pop(0)  # ['Q_OBJECT']
                var = tokens.pop() if tokens else None
                if var == 'Q_OBJECT':
                    result.q_object = True
                    continue

            if each.cursor.kind == clang.cindex.CursorKind.CXX_ACCESS_SPEC_DECL:
                access = _access_spec_decl_from_node(each)

            MEMBER_TYPES = {
                clang.cindex.CursorKind.CXX_METHOD,
                clang.cindex.CursorKind.TYPE_ALIAS_TEMPLATE_DECL,
                clang.cindex.CursorKind.TYPE_ALIAS_DECL,
            }
            if each.cursor.kind in MEMBER_TYPES:
                member = from_node(each)
                member.access = access
                result.members.append(member)

        return result


@dataclasses.dataclass
class Variable:
    """For C++ member variable declarations.

    Attributes:
        name: The variable name
        type: The type declaration
        default_args: The default value of the member
        mutable: Indicates if the member is mutable
        access: The member's access specifier
    """
    name: str
    type: Type
    default_args: Sequence[str] = dataclasses.field(default_factory=list)
    mutable: bool = False
    access: str = 'public'

    def __str__(self):
        result = ''
        result += self.mutable * 'mutable '
        result += str(self.type) + ' '
        result += self.name
        result += '{' + ', '.join(self.default_args) + '}'
        result += ';'
        return result


def _access_spec_decl_from_node(node: translator.Node) -> str:
    """Get access specifier token from ``ACCESS_SPEC_DECL``."""
    tokens = node.get_tokens()  # public, protected, private (slots)
    # Since `slots` and `signals` is #defined as empty, an empty access
    # spec decl is most likely ``signals:``.
    if not tokens:
        return 'signals'
    return tokens[0]



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any, Sequence

import re

INDENT_WIDTH = 2


def template(params: Sequence[Any]) -> str:
    """Join sequence in angled braces."""
    return '<' + ', '.join(str(each) for each in params) + '>'


def swap(regex: str, dst: str, src: str) -> str:
    """Capture ``src`` with ``regex`` and replace every substring
    `r'\1'` in ``dst`` with the content of the first capture group.

    Raises:
        IndexError if ``src`` doesn't match ``regex`` and
        ValueError if ``regex`` doesn't match ``src``.

    Example:
        >>> regex = r'[0-9]([a-z])foo'
        >>> dst = r'foo\1'
        >>> x = '4barfooo'
        >>> swap(regex, dst, x)  # "foobar"
    """
    if r'\1' not in dst:
        return dst

    # Find the capture group's match; raise if there is no match.
    match = re.match(regex, src)
    if not match:
        raise ValueError(f'{src} doesn\'t match {regex}.')
    inner = match.group(1)
    result = dst.replace(r'\1', inner)
    return result


def split_by_condition(pred, seq: Sequence[_T]) -> list[list[_T]]:
    """Split ``seq`` into lists of equivalence classes.

    Args:
        pred:
            The predicate which sorts the sequence into equivalence
            classes
        seq:
            The sequence to split

    Returns:
        A list which contains the equivalence classes as lists
    """
    # Get values in order of occurence.
    values = []
    for each in seq:
        val = pred(each)
        if val not in values:
            values.append(val)
    return [[each for each in seq if pred(each) == val] for val in values]


def indent(value: str, depth: int = 1, width: int = INDENT_WIDTH) -> str:
    """Indent a string according to depth.

    Args:
        value: The string to indent
        depth: The depth of the string (number of tabs/indents)
        width: Indent width
    """
    result = value
    result = depth * width * ' ' + result
    result = result.replace('\n', '\n' + depth * width * ' ')
    return result


class DrMockRuntimeError(Exception):
    pass



# SPDX-FileCopyrightText: 2021 Malte Kliemann, Ole Kliemann
#
# SPDX-License-Identifier: GPL-3.0-or-later

from setuptools import setup


setup(
    name='drmock-generator',
    author='Malte Kliemann, Ole Kliemann',
    version='0.6.0-beta',
    packages=['drmock'],
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'drmock-gen = drmock.commandline:main'
        ]
    },
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=[
        'clang>=11.0',
    ]
)
